from .segment_tree import SegmentTree


class MinSegmentTree(SegmentTree):
    def __init__(self, capacity: int):
        super(MinSegmentTree, self).__init__(capacity=capacity, operation=min, init_value=float("inf"))

    def min(self, start: int = 0, end: int = 0) -> float:
        return super(MinSegmentTree, self).operate(start, end)



import typing as T


class EvolvingInt:
    def __init__(self, value: int, min_limit: int, max_limit: int, mutation_step: float):
        if value > max_limit:
            value = max_limit
        elif value < min_limit:
            value = min_limit
        self.value: int = value
        self.min_limit: int = min_limit
        self.max_limit: int = max_limit
        self.mutation_step: float = mutation_step


class EvolvingFloat:
    def __init__(self, value: float, min_limit: float, max_limit: float, mutation_step: float):
        if value > max_limit:
            value = max_limit
        elif value < min_limit:
            value = min_limit
        self.value: float = value
        self.min_limit: float = min_limit
        self.max_limit: float = max_limit
        self.mutation_step: float = mutation_step


class EvolvingBool:
    def __init__(self, value: bool, mutation_step: float):
        self.value: float = value
        self.mutation_step: float = mutation_step


class EvolutionerParams:
    def __init__(self, workers: int, generation_size: int):
        self.workers: int = workers
        self.generation_size: int = generation_size


T_EParams = T.Dict[str, T.Union[EvolvingInt, EvolvingBool, EvolvingFloat]]


class EvolutionProgress:
    def __init__(self, results: T.List[float], params: T.List[T_EParams], generation: int, best_index: int):
        self.best_index: int = best_index
        self.results: T.List[float] = results
        self.params: T.List[T_EParams] = params
        self.generation: int = generation



import logging
import os
import sys
from abc import ABC


class BaseObject(ABC):
    def __init__(self):
        self.log: logging.Logger = logging.getLogger(self.__class__.__name__)
        handler = logging.StreamHandler(stream=sys.stdout)
        formatter = logging.Formatter('%(asctime)s %(module)-20s %(levelname)-5s %(message)s')
        handler.setFormatter(formatter)
        if not self.log.hasHandlers():
            self.log.addHandler(handler)
        self.log.setLevel(self._get_debug_level())

    def _get_debug_level(self) -> int:
        level = logging.ERROR
        debug_info = os.getenv('LZ_DEBUG', '0')
        for debug_element_level in debug_info.split(","):
            split_debug_element_level = debug_element_level.split(":")
            if len(split_debug_element_level) == 2:
                element_name = split_debug_element_level[0]
                debug_level_str = split_debug_element_level[1]
                if element_name == self.__class__.__name__ and debug_level_str.isnumeric():
                    debug_level = int(debug_level_str)
                else:
                    continue

            elif len(split_debug_element_level) == 1 and split_debug_element_level[0].isnumeric():
                debug_level = int(split_debug_element_level[0])

            else:
                print(f"incorrect debug specification {debug_element_level}")
                continue

            if debug_level == 0:
                level = logging.ERROR
            elif debug_level == 1:
                level = logging.WARNING
            elif debug_level == 2:
                level = logging.INFO
            elif debug_level == 3:
                level = logging.DEBUG
        return level



import typing as T

import gym
import numpy as np

from .environment import Environment


class CartPole(Environment):
    OBSERVATION_SPACE = 4
    ACTION_SPACE = 2

    def __init__(self):
        self.visualize: bool = True
        self.env: gym.Env = gym.make("CartPole-v1")

    def reset(self) -> np.ndarray:
        return self.env.reset()

    def step(self, action: int) -> T.Tuple[np.ndarray, float, bool]:
        s, r, f, _ = self.env.step(action)
        return s, r, f

    def render(self) -> None:
        if self.visualize:
            self.env.render()

    def close(self) -> None:
        self.env.close()



import asyncio
import random
import typing as T
from abc import ABC, abstractmethod

import numpy as np
import torch

from .individual import Individual
from .models import EvolutionProgress, EvolutionerParams, T_EParams, EvolvingFloat, EvolvingBool, EvolvingInt
from ..agents import AnyAgent
from ..environments import Environment


class Evolutioner(ABC):
    def __init__(self,
                 env: Environment,
                 init_params: T_EParams,
                 evolutioner_params: EvolutionerParams):

        self.env: Environment = env
        self.evolutioner_params: EvolutionerParams = evolutioner_params
        self.init_params: T_EParams = init_params
        self.progress_callback: T.Callable[[EvolutionProgress], None] = lambda x: None
        self.fitness_function: T.Callable[[np.ndarray], float] = lambda x: np.mean(x).item()

    @abstractmethod
    def agent_factory(self, params: T_EParams, state_dict: T.Optional[dict]) -> AnyAgent:
        raise NotImplementedError()

    def set_progress_callback(self, cbk: T.Callable[[EvolutionProgress], None]) -> None:
        self.progress_callback = cbk

    def create_generation(self, params: T_EParams, state_dict: dict) -> T.Tuple[T.List[Individual], T.List[T_EParams]]:
        result: T.List[Individual] = []
        params_result: T.List[T_EParams] = []
        size = self.evolutioner_params.generation_size
        for i in range(size):
            mutated_params = self.mutate(params, i/size)
            agent = self.agent_factory(mutated_params, state_dict)
            result.append(Individual(agent, self.env))
            params_result.append(mutated_params)
        return result, params_result

    @staticmethod
    def mutate(params: T_EParams, scale_mutation: float) -> T_EParams:
        result = {}
        for k, v in params.items():
            if isinstance(v, EvolvingFloat):
                v_dict = v.__dict__
                mutation = (1 if random.random() > 0.5 else -1) * scale_mutation * v.mutation_step * random.random()
                v_dict["value"] += mutation
                result[k] = EvolvingFloat(**v_dict)
            elif isinstance(v, EvolvingInt):
                v_dict = v.__dict__
                mutation = (1 if random.random() > 0.5 else -1) * scale_mutation * v.mutation_step * random.random()
                v_dict["value"] = int(mutation + v_dict["value"])
                result[k] = EvolvingInt(**v_dict)
            elif isinstance(v, EvolvingBool):
                v_dict = v.__dict__
                mutation = scale_mutation * v.mutation_step
                if random.random() > mutation:
                    v_dict["value"] = not v_dict["value"]
                result[k] = EvolvingBool(**v_dict)
            else:
                raise ValueError(f"value from param {k} has type {type(v)}. Only types int, float and bool are valid")
        return result

    def evolve(self, finish_condition: T.Callable[[EvolutionProgress], bool]) -> None:
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._evolve(finish_condition))

    async def _evolve(self, finish_condition: T.Callable[[EvolutionProgress], bool]) -> None:
        torch.multiprocessing.set_start_method("spawn")
        params = self.init_params.copy()
        next_state_dict = None
        generation_count = 0
        while True:
            generation, mutated_params = self.create_generation(params, next_state_dict)
            results: T.List[T.Union[float, None]] = [None for _ in range(len(generation))]
            state_dicts: T.List[T.Union[dict, None]] = [None for _ in range(len(generation))]

            tasks: T.List[T.Coroutine] = []
            completed = []
            for i in range(len(generation)):
                async def task(index: int):
                    life_result, state_dict = await generation[index].parallel_life(index)
                    results[index] = self.fitness_function(life_result)
                    state_dicts[index] = state_dict
                    completed.append(None)
                    print("=============================")
                    print(f"completed {len(completed)}/{len(generation)}: {results[index]}")
                    for k, v in mutated_params[index].items():
                        print(k+":", v.value)

                tasks.append(task(i))
                if i >= len(generation) - 1:
                    await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
                    break
                if len(tasks) >= self.evolutioner_params.workers:
                    _, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
                    tasks = list(pending)

            best_index = np.argmax(results).item()
            progress = EvolutionProgress(
                results,
                mutated_params,
                generation_count,
                best_index
            )

            self.progress_callback(progress)
            if finish_condition(progress):
                break
            generation_count += 1
            params = mutated_params[best_index]
            next_state_dict = state_dicts[best_index]



import numpy as np

from . import tools
from .test_random_replay_buffer import is_random
from ..base.n_steps_replay_buffer import NStepsReplayBuffer
from ...replay_buffers import NStepRandomReplayBufferParams, NStepsRandomReplayBuffer
from ...base.models import ReplayBufferEntry


def is_n_step(replay_buffer: NStepsReplayBuffer):
    n_step, gamma = replay_buffer.rp.n_step, replay_buffer.agent_params.gamma
    final_step = 6
    records = []
    for i in range(14):
        rpe = ReplayBufferEntry(np.array([i]), np.array([i]), i, i, i == final_step)
        records.append(rpe)
        replay_buffer.rp_add(rpe)

    assert replay_buffer.rp_get_length() == 14 - n_step + 1

    samples = replay_buffer.rp_sample(8)
    for sample in samples:
        i = sample.a
        rest = 0
        if final_step - n_step + 1 <= i <= final_step:
            rest = i - (final_step - n_step + 1)
        assert sample.s_.item() == records[i + n_step - 1 - rest].s_.item()
        assert round(sample.r, 3) == round(sum([ii * (gamma ** n) for n, ii in enumerate(range(i, i + n_step - rest))]), 3)


class Agent(NStepsRandomReplayBuffer, tools.Agent):
    pass


def test_elements_are_n_step():
    nrp = NStepRandomReplayBufferParams(max_len=15, n_step=3)
    test_agent = Agent(replay_buffer_params=nrp)
    is_n_step(test_agent)


def test_elements_retrieved_are_random():
    nrp = NStepRandomReplayBufferParams(max_len=10, n_step=1)
    test_agent = Agent(replay_buffer_params=nrp)
    is_random(test_agent)



import typing as T
from .random_explorer import RandomExplorer
from .noisy_explorer import NoisyExplorer
from .base.params import RandomExplorerParams, NoisyExplorerParams
AnyExplorer = T.Union[RandomExplorer, NoisyExplorer]



import typing as T
from dataclasses import dataclass, field


@dataclass
class RandomExplorerParams:
    init_ep: float = 1.0
    final_ep: float = 0.01
    decay_ep: float = 1e-3


@dataclass
class NoisyExplorerParams:
    extra_layers: T.List[int] = field(default_factory=list)
    reset_noise_every: int = 1
    std_init: float = 0.5



import typing as T
from abc import ABC

import torch

from . import models, base_object
from .. import interfaces
from ...environments import Environment


class BaseAgent(base_object.BaseObject,
                interfaces.AgentEventInterface,
                interfaces.ExplorerInterface,
                interfaces.ReplayBufferInterface,
                ABC):
    def __init__(self, action_space: int, agent_params: models.AgentParams, use_gpu: bool = True):
        super(BaseAgent, self).__init__()
        self.device: str = "cpu"
        if use_gpu:
            if not torch.cuda.is_available():
                self.log.warning("cuda is not available, CPU will be used")
            else:
                self.device = "cuda"
        self.action_space = action_space
        self.agent_params = agent_params
        self.default_training_params = models.TrainingParams()
        self.step_callbacks: T.Dict[str, models.TStepCallback] = {}
        self.progress_callbacks: T.Dict[str, models.TProgressCallback] = {}
        self.learning_callbacks: T.Dict[str, models.TLearnCallback] = {}
        self.accumulate_rewards: bool = True
        self.rp_link()
        self.ex_link()

    def get_self_class_name(self):
        return self.__class__.__name__

    @staticmethod
    def last_layer_factory(in_features: int, out_features: int) -> torch.nn.Module:
        return torch.nn.Linear(in_features, out_features)

    def health_check(self, env: Environment) -> None:
        self.log.info("checking the model is healthy...")
        s = env.reset()
        self.log.debug(f"state for testing health is:\n{s}")
        self.log.info("testing preprocessing...")
        try:
            self.preprocess(s)
        except Exception as e:
            self.log.error("error while testing preprocessing")
            raise e
        self.log.info("preprocessing is correct")
        self.log.info("testing inference...")
        try:
            self.act(s)
        except Exception as e:
            self.log.error("error while testing inference")
            raise e
        self.log.info("inference is correct")
        self.log.info("testing learning...")
        while self.rp_get_length() < 2:
            a = 0
            s_, r, final = env.step(a)
            self.rp_add(models.ReplayBufferEntry(s, s_, a, r, final))
            s = s_
            if final:
                s = env.reset()

        batch = self.rp_sample(2)
        try:
            self.learn(batch)
        except Exception as e:
            self.log.error("error while testing learning")
            raise e

        self.log.info("learning is correct")
        self.log.info("model is healthy!")
        self.rp_clear()

    def add_step_callback(self, label: str, cbk: models.TStepCallback) -> None:
        if label in self.step_callbacks:
            self.log.warning(f"overwriting step callback with label {label}")
        self.step_callbacks[label] = cbk
        self.log.debug(f"added step callback {label}: {cbk}")

    def add_progress_callback(self, label: str, cbk: models.TProgressCallback) -> None:
        if label in self.progress_callbacks:
            self.log.warning(f"overwriting progress callback with label {label}")
        self.progress_callbacks[label] = cbk
        self.log.debug(f"added progress callback {label}: {cbk}")

    def add_learn_callback(self, label: str, cbk: models.TLearnCallback) -> None:
        if label in self.learning_callbacks:
            self.log.warning(f"overwriting learning callback with label {label}")
        self.learning_callbacks[label] = cbk
        self.log.debug(f"added learning callback {label}: {cbk}")

    def call_step_callbacks(self, training_step: models.TrainingStep) -> None:
        self.log.debug("calling step callbacks...")
        for label, cbk in self.step_callbacks.items():
            self.log.debug(f"calling step callback {label}, {cbk}")
            cbk(training_step)
        self.log.debug("all step callbacks called")

    def call_progress_callbacks(self, training_progress: models.TrainingProgress) -> bool:
        self.log.debug("calling progress callbacks...")
        must_exit = False
        for label, cbk in self.progress_callbacks.items():
            self.log.debug(f"calling progress callback {label}, {cbk}")
            may_exit = cbk(training_progress)
            if may_exit:
                must_exit = True
                self.log.warning(f"progress callback {label} said that training should end")
        self.log.debug("all progress callbacks called")
        return must_exit

    def call_learn_callbacks(self, learning_step: models.LearningStep) -> None:
        self.log.debug("calling learning callbacks...")
        for label, cbk in self.learning_callbacks.items():
            self.log.debug(f"calling learning callback {label}, {cbk}")
            cbk(learning_step)
        self.log.debug("all learning callbacks called")

    def forward_hook(self, module: torch.nn.Module, x: T.Tuple[torch.Tensor, ...], y: torch.Tensor) -> None:
        pass

    def build_model(self) -> torch.nn.Module:
        self.log.info("building model from model factory...")
        model = self.model_factory()
        model.register_forward_hook(self.forward_hook)
        self.log.info("model built correctly, applying last layers modifier...")
        model = self.last_layers_model_modifier(model)
        self.log.info("last layers modified correctly, applying agent specification")
        model = self.agent_specification_model_modifier(model)
        self.log.info("agent specification applied correctly")
        return model

    def act(self, x):
        with torch.no_grad():
            preprocessed = self.preprocess(x)
            if type(preprocessed) == tuple:
                preprocessed = tuple(p.unsqueeze(0).to(self.device) for p in preprocessed)
            else:
                preprocessed = preprocessed.unsqueeze(0).to(self.device)
            infer = self.infer(preprocessed)
            return self.postprocess(infer.cpu())

    def form_learning_batch(self, batch: T.List[models.ReplayBufferEntry]) -> models.LearningBatch:
        preprocessed_sample = self.preprocess(batch[0].s)
        if type(preprocessed_sample) == tuple:
            p_batch_s = [self.preprocess(entry.s) for entry in batch]
            p_batch_s_ = [self.preprocess(entry.s_) for entry in batch]
            batch_s = tuple(torch.stack([p_s[i] for p_s in p_batch_s], 0).to(self.device).requires_grad_(True) for i in range(len(preprocessed_sample)))
            batch_s_ = tuple(torch.stack([p_s_[i] for p_s_ in p_batch_s_], 0).to(self.device) for i in range(len(preprocessed_sample)))
        else:
            batch_s = torch.stack([self.preprocess(m.s) for m in batch], 0).to(self.device).requires_grad_(True)
            batch_s_ = torch.stack([self.preprocess(m.s_) for m in batch], 0).to(self.device)
        batch_a = torch.tensor([m.a for m in batch], device=self.device)
        batch_r = torch.tensor([m.r for m in batch], dtype=torch.float32, device=self.device)
        batch_finals = torch.tensor([int(not m.final) for m in batch], device=self.device)
        batch_weights = torch.tensor([m.weight for m in batch], device=self.device)
        return models.LearningBatch(batch_s, batch_s_, batch_a, batch_r, batch_finals, batch_weights)



import typing as T

import gym
import numpy as np

from .environment import Environment


class FrozenLake(Environment):
    def __init__(self):
        self.visualize: bool = True
        self.env: gym.Env = gym.make("FrozenLake-v0")

    @staticmethod
    def form_state(index: int) -> np.ndarray:
        s = np.zeros(16)
        s[index] = 1
        return s

    def reset(self) -> np.ndarray:
        return self.form_state(self.env.reset())

    def step(self, action: int) -> T.Tuple[np.ndarray, float, bool]:
        s, r, f, _ = self.env.step(action)
        return self.form_state(s), 1 if r > 0 else (-1 if f else 0), f

    def render(self) -> None:
        if self.visualize:
            self.env.render()

    def close(self) -> None:
        self.env.close()



from abc import ABC

from .base.models import DoubleDuelingDqnHyperParams
from .double_dqn import DoubleDqnAgent
from .dueling_dqn import DuelingDqnAgent


class DoubleDuelingDqnAgent(DuelingDqnAgent, DoubleDqnAgent, ABC):
    def __init__(self,
                 action_space: int,
                 agent_params: DoubleDuelingDqnHyperParams = DoubleDuelingDqnHyperParams(),
                 use_gpu: bool = True):
        super(DoubleDuelingDqnAgent, self).__init__(action_space, agent_params, use_gpu)



import math
import typing as T
from abc import ABC

import numpy as np
import torch
import torch.nn.functional as F

from .base.explorer import Explorer
from .base.params import NoisyExplorerParams


class NoisyLinear(torch.nn.Module):
    def __init__(self, in_features: int, out_features: int, std_init: float):
        super(NoisyLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.std_init = std_init
        self.weight_mu = torch.nn.Parameter(torch.empty(out_features, in_features))
        self.weight_sigma = torch.nn.Parameter(torch.empty(out_features, in_features))
        self.register_buffer('weight_epsilon', torch.empty(out_features, in_features))
        self.bias_mu = torch.nn.Parameter(torch.empty(out_features))
        self.bias_sigma = torch.nn.Parameter(torch.empty(out_features))
        self.register_buffer('bias_epsilon', torch.empty(out_features))
        self.reset_parameters()
        self.reset_noise()

    def reset_parameters(self) -> None:
        mu_range = 1 / math.sqrt(self.in_features)
        self.weight_mu.data.uniform_(-mu_range, mu_range)
        self.weight_sigma.data.fill_(self.std_init / math.sqrt(self.in_features))
        self.bias_mu.data.uniform_(-mu_range, mu_range)
        self.bias_sigma.data.fill_(self.std_init / math.sqrt(self.out_features))

    def _scale_noise(self, size: int) -> torch.Tensor:
        x = torch.FloatTensor(np.random.normal(loc=0.0, scale=1.0, size=size)).to(self.weight_mu.device)
        return x.sign().mul_(x.abs().sqrt_())

    def reset_noise(self) -> None:
        epsilon_in = self._scale_noise(self.in_features)
        epsilon_out = self._scale_noise(self.out_features)
        self.weight_epsilon.copy_(epsilon_out.ger(epsilon_in))
        self.bias_epsilon.copy_(epsilon_out)

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}'.format(self.in_features, self.out_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.linear(x,
                        self.weight_mu + self.weight_sigma * self.weight_epsilon,
                        self.bias_mu + self.bias_sigma * self.bias_epsilon)


class ModelWithNoisyLayers(torch.nn.Module):
    def __init__(self, model: torch.nn.Module, noisy_layers: T.List[torch.nn.Module]):
        super(ModelWithNoisyLayers, self).__init__()
        self.model = model
        self.noisy_layers = torch.nn.Sequential(*noisy_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)
        x = self.noisy_layers(x)
        return x


class NoisyExplorer(Explorer, ABC):
    def __init__(self, explorer_params: NoisyExplorerParams = NoisyExplorerParams(), *args, **kwargs):
        if not isinstance(explorer_params, NoisyExplorerParams):
            raise ValueError("argument ep must be an instance of NoisyExplorerParams")
        self.ep: NoisyExplorerParams = explorer_params
        self.noisy_layers_reference: T.List[NoisyLinear] = []
        self.reset_count: int = 0
        super(NoisyExplorer, self).__init__(*args, **kwargs)

    def last_layers_model_modifier(self, model: torch.nn.Module) -> torch.nn.Module:
        self.log.info(f"wrapping model with noisy layers")
        last_layer = list(model.modules())[-1]
        if not isinstance(last_layer, torch.nn.Linear):
            raise ValueError("the model you have created must have a torch.nn.Linear in the last layer")
        if len(self.ep.extra_layers) == 0:
            return model
        noisy_layers = [NoisyLinear(last_layer.out_features, self.ep.extra_layers[0], self.ep.std_init)]
        for i in range(1, len(self.ep.extra_layers)):
            noisy_layers.append(torch.nn.ReLU())
            noisy_layers.append(NoisyLinear(self.ep.extra_layers[i - 1], self.ep.extra_layers[i], self.ep.std_init))

        self.noisy_layers_reference += noisy_layers
        return ModelWithNoisyLayers(model, noisy_layers)

    def ex_choose(self, actions: np.ndarray, f: T.Callable[[np.ndarray], int]) -> int:
        return f(actions)

    def reset_noise(self, *_, **__):
        self.reset_count += 1
        if self.reset_count < self.ep.reset_noise_every:
            return
        self.reset_count = 0
        i = 0
        for layer in self.noisy_layers_reference:
            if isinstance(layer, NoisyLinear):
                self.log.debug(f"resetting noise for noise layer {i}")
                layer.reset_noise()
                i += 1

    def last_layer_factory(self, in_features: int, out_features: int) -> NoisyLinear:
        noisy_linear = NoisyLinear(in_features, out_features, self.ep.std_init)
        self.noisy_layers_reference.append(noisy_linear)
        return noisy_linear

    def ex_link(self):
        self.log.info(f"linking {type(self).__name__}...")
        self.add_step_callback("noisy explorer reset noisy", self.reset_noise)

    def ex_get_stats(self) -> T.Dict[str, float]:
        return {}



from ..noisy_explorer import NoisyExplorer, NoisyExplorerParams, NoisyLinear
import random
import torch
from . import tools


class Agent(NoisyExplorer, tools.Agent):
    pass


def test_generated_model_works():
    nep = NoisyExplorerParams(reset_noise_every=1, std_init=0.5, extra_layers=[2, 1])

    class Model(torch.nn.Module):
        def __init__(self):
            super(Model, self).__init__()
            self.linear = torch.nn.Linear(2, 2)

        def forward(self, x):
            return self.linear(x)

    test_agent = Agent(explorer_params=nep)
    model = test_agent.last_layers_model_modifier(Model())
    assert len([l for l in model.modules() if isinstance(l, torch.nn.Linear)]) == 1
    assert len([l for l in model.modules() if isinstance(l, NoisyLinear)]) == 2

    op = torch.optim.Adam(model.parameters(), lr=0.005)

    bs, ep = 8, 6000
    x = torch.zeros((bs, 2), dtype=torch.float32)
    y = torch.zeros((bs, 1), dtype=torch.float32)
    for i in range(ep):
        for j in range(bs):
            a, b = 100*random.random(), 100*random.random()
            x[j, 0], x[j, 1] = a, b
            y[j][0] = a + b
        y_ = model(x)
        loss = (y - y_).pow(2).mean()
        op.zero_grad()
        loss.backward()
        op.step()
        if i % nep.reset_noise_every == 0:
            [l.reset_noise() for l in model.modules() if isinstance(l, NoisyLinear)]

    result = model(torch.Tensor([[2., 3.]])).squeeze(0).item()
    assert 4.5 < result < 5.5
    [l.reset_noise() for l in model.modules() if isinstance(l, NoisyLinear)]
    result = model(torch.Tensor([[2., 3.]])).squeeze(0).item()
    assert 4.5 < result < 5.5



from .... import agents
from abc import ABC
import torch
import numpy as np


class Model(torch.nn.Module):
    def __init__(self):
        super(Model, self).__init__()
        self.linear = torch.nn.Linear(2, 2)

    def forward(self, x):
        return self.linear(x)


class Agent(agents.replay_buffers.RandomReplayBuffer, agents.DqnAgent, ABC):
    def __init__(self, *args, **kwargs):
        super(Agent, self).__init__(*args, action_space=1, **kwargs)

    def model_factory(self) -> torch.nn.Module:
        return Model()

    def preprocess(self, x: np.ndarray) -> torch.Tensor:
        return torch.tensor(x)



import typing
from .a2c import A2cAgent, A2CHyperParams
from .monte_carlo_advantage_actor_critic import MonteCarloA2c
from .dqn import DqnAgent, DqnHyperParams
from .double_dqn import DoubleDqnAgent, DoubleDqnHyperParams
from .double_dueling_dqn import DoubleDuelingDqnAgent, DoubleDuelingDqnHyperParams
from .dueling_dqn import DuelingDqnAgent, DuelingDqnHyperParams
from .ppo import PpoAgent, PpoHyperParams
from .base.models import TrainingStep, TrainingParams, TrainingProgress, LearningStep
from . import replay_buffers, explorers, loggers

AnyAgent = typing.Union[DqnAgent, DoubleDqnAgent, DuelingDqnAgent, DoubleDuelingDqnAgent, A2cAgent, MonteCarloA2c, PpoAgent]



from dataclasses import dataclass


@dataclass
class ReplayBufferParams:
    max_len: int = 10000


@dataclass
class RandomReplayBufferParams(ReplayBufferParams):
    pass


@dataclass
class NStepReplayBufferParams(ReplayBufferParams):
    n_step: int = 3


@dataclass
class PrioritizedReplayBufferParams(ReplayBufferParams):
    alpha: float = 0.6
    init_beta: float = 0.4
    final_beta: float = 1.0
    increase_beta: float = 1e-4


@dataclass
class NStepPrioritizedReplayBufferParams(NStepReplayBufferParams, PrioritizedReplayBufferParams):
    pass


@dataclass
class NStepRandomReplayBufferParams(NStepReplayBufferParams, RandomReplayBufferParams):
    pass



import typing as T
from abc import ABC

import numpy as np
import torch

from .a2c import A2cAgent
from .base.models import TrainingProgress, LearningStep, TrainingStep, TrainingParams, ReplayBufferEntry
from ..environments import Environment


class MonteCarloA2c(A2cAgent, ABC):
    def learn(self, entries: T.List[ReplayBufferEntry]) -> None:
        batch = self.form_learning_batch(entries)
        batch.r = batch.r.unsqueeze(1)

        action_probabilities, state_values = self.actor_critic(batch.s)

        advantages: torch.Tensor = (batch.r - state_values.clone().detach()).squeeze(1)
        chosen_action_log_probabilities: torch.Tensor = torch.stack(
            [torch.distributions.Categorical(p).log_prob(a) for p, a in zip(action_probabilities, batch.a)])
        actor_loss: torch.Tensor = -chosen_action_log_probabilities * advantages * batch.weight
        critic_loss: torch.Tensor = self.loss_f(state_values, batch.r) * batch.weight
        loss = (actor_loss + critic_loss).mean()
        self.actor_critic_optimizer.zero_grad()
        loss.backward()
        self.actor_critic_optimizer.step()
        self.call_learn_callbacks(LearningStep(entries, [v.item() for v in state_values], [v.item() for v in batch.r]))

    def train(self, env: Environment, tp: TrainingParams = None) -> None:
        if tp is None:
            tp = self.default_training_params
        self.health_check(env)
        self.accumulate_rewards = False

        s = env.reset()
        i = 0
        episode = 1
        steps_survived = 0
        accumulated_reward = 0
        steps_record: T.List[ReplayBufferEntry] = []
        while True:
            estimated_rewards = self.act(s)
            def choosing_f(x): return torch.distributions.Categorical(torch.tensor(x)).sample().item()
            a = self.ex_choose(list(estimated_rewards), choosing_f)
            s_, r, final = env.step(a)
            steps_record.append(ReplayBufferEntry(s, s_, a, r, final))
            accumulated_reward += r
            s = s_

            self.call_step_callbacks(TrainingStep(i, episode))

            if i % self.agent_params.learn_every == 0 and i != 0 and self.rp_get_length() >= tp.batch_size:
                batch = self.rp_sample(tp.batch_size)
                self.learn(batch)

            if final:
                discounted_r = 0
                reward_array = np.zeros((len(steps_record)))
                for j, step in enumerate(steps_record[::-1]):
                    discounted_r = step.r + self.agent_params.gamma * discounted_r
                    step.r = discounted_r
                    reward_array[j] = discounted_r

                mean, std, eps = reward_array.mean(), reward_array.std(), np.finfo(np.float32).eps.item()
                for step in steps_record:
                    step.r = (step.r - mean) / (std + eps)
                    self.rp_add(step)

                training_progress = TrainingProgress(i, episode, steps_survived, accumulated_reward)
                must_exit = self.call_progress_callbacks(training_progress)
                if episode >= tp.episodes or must_exit:
                    return

                accumulated_reward = 0
                steps_survived = 0
                episode += 1
                steps_record.clear()
                s = env.reset()

            else:
                steps_survived += 1
            env.render()
            i += 1



import datetime
import json
import os
import typing as T
from abc import ABC

import torch

from .tensorboard_logger import TensorBoard
from ..base import models
from ..base.base_agent import BaseAgent
from ...environments import Environment


DISABLE_EMBEDDING = True


class TensorBoardLogger(BaseAgent, ABC):
    def __init__(self, *args, **kwargs):
        self.summary_writer: T.Union[None, TensorBoard] = None
        self.sample_inputs: T.Union[None, T.List[torch.Tensor]] = None
        self.module_names: T.List[str] = []
        self.modules: T.Dict[torch.nn.Module, T.Dict] = {}
        super().__init__(*args, **kwargs)
        self.link_tensorboard()

    def __del__(self):
        if self.summary_writer:
            self.summary_writer.__del__()

    def health_check(self, env: Environment):
        super(TensorBoardLogger, self).health_check(env)
        folder = self.create_tensor_board_folder(env)
        self.summary_writer = TensorBoard(folder)
        self.tensorboard_log_model_graph()

    def tensorboard_log_training_progress(self, training_progress: models.TrainingProgress) -> bool:
        if self.summary_writer:
            self.summary_writer.add_scalar("episode reward", training_progress.total_reward, training_progress.episode)

        return False

    def tensorboard_log_hyper_params(self):
        info = self.get_info()
        result = {}

        def nested(root: T.Any, acc: str = ""):
            if not isinstance(root, dict):
                result[acc] = json.dumps(root) if isinstance(root, tuple) else root
            else:
                for key in root:
                    nested(root[key], acc + " -> " + key if acc != "" else key)

        nested(info)
        for r in result:
            self.summary_writer.add_text(r, str(result[r]))

    def tensorboard_log_model_graph(self):
        models_in_self: T.Dict[str, torch.nn.Module] = {}
        for attr, value in self.__dict__.items():
            if isinstance(value, torch.nn.Module) and not attr.startswith("loss") and self.sample_inputs is not None:
                models_in_self[attr] = value

        class AllModels(torch.nn.Module):
            def __init__(self):
                super(AllModels, self).__init__()
                for name, model in models_in_self.items():
                    self.__setattr__(name, model)

            def forward(self, x):
                result_unfolded = []
                for result in [self.__getattr__(name)(x) for name in models_in_self]:
                    if isinstance(result, tuple):
                        for folded_result in result:
                            result_unfolded.append(folded_result)
                    else:
                        result_unfolded.append(result)
                return tuple(result_unfolded)

        self.summary_writer.add_graph(AllModels(), self.sample_inputs)

    def tensorboard_log_explorer_stats_progress_callback(self, training_progress: models.TrainingProgress) -> bool:
        if self.summary_writer:
            for k, v in self.ex_get_stats().items():
                self.summary_writer.add_scalar(k, v, training_progress.episode)
        return False

    def tensorboard_log_replay_buffer_stats_progress_callback(self, training_progress: models.TrainingProgress) -> bool:
        if self.summary_writer:
            for k, v in self.rp_get_stats().items():
                self.summary_writer.add_scalar(k, v, training_progress.episode)
        return False

    def forward_hook(self, module: torch.nn.Module, x: T.Tuple[torch.Tensor], y: torch.Tensor) -> None:
        if self.sample_inputs is None:
            self.sample_inputs = x
        if DISABLE_EMBEDDING:
            return
        if self.summary_writer:
            if len(self.module_names) == 0:
                for attr, value in self.__dict__.items():
                    if isinstance(value, torch.nn.Module) and not attr.startswith("loss"):
                        self.module_names.append(attr)
            if module not in self.modules:
                self.modules[module] = {"name": self.module_names[len(self.modules)], "times": 0, "renders": 0}

            if self.modules[module]["times"] % 1000 == 0:
                self.summary_writer.add_embedding(y,
                                                  tag=self.modules[module]["name"],
                                                  global_step=self.modules[module]["renders"])
                self.modules[module]["renders"] += 1
            self.modules[module]["times"] += 1

    def link_tensorboard(self) -> None:
        self.log.info("linking tensorboard callbacks...")
        self.add_progress_callback("tensorboard log training progress",
                                   self.tensorboard_log_training_progress)
        self.add_progress_callback("tensorboard log explorer stats",
                                   self.tensorboard_log_explorer_stats_progress_callback)
        self.add_progress_callback("tensorboard log replay buffer stats",
                                   self.tensorboard_log_replay_buffer_stats_progress_callback)

    def create_tensor_board_folder(self, env: Environment) -> str:
        base = os.environ.get("LZ_TENSORBOARD_PATH", "data")
        if base.endswith("/"):
            base = base[:-1]

        agent = self.get_self_class_name()
        today = str(datetime.datetime.now().date())
        now = str(datetime.datetime.now().time().strftime("%H_%M_%S"))
        folder = ""
        for sub_folder in [base, agent, type(env).__name__, today, now]:
            folder = os.path.join(folder, sub_folder)
            if not os.path.isdir(folder):
                self.log.info(f"folder {folder} does not exists, creating it...")
                os.mkdir(folder)
        self.log.info(f"all save folders created: {folder}")
        return folder



import numpy as np

from . import tools
from ...replay_buffers import RandomReplayBuffer, RandomReplayBufferParams
from ...base.models import ReplayBufferEntry


def is_random(replay_buffer: RandomReplayBuffer):
    for i in range(14):
        replay_buffer.rp_add(ReplayBufferEntry(np.array([]), np.array([]), i, i, False))

    samples = replay_buffer.rp_sample(5)
    for i, sample in enumerate(samples, 9):
        if sample.a != i:
            break
    else:
        raise AssertionError()

    indexes = [s.a for s in samples]
    assert indexes != sorted(indexes)
    assert len(indexes) == len(set(indexes))


class Agent(RandomReplayBuffer, tools.Agent):
    pass


def test_elements_retrieved_are_random():
    lrp = RandomReplayBufferParams(max_len=10)
    test_agent = Agent(replay_buffer_params=lrp)
    is_random(test_agent)



import typing as T
from abc import ABC, abstractmethod

import torch

from ..base import models
from ...environments import Environment


class AgentEventInterface(ABC):
    @abstractmethod
    def health_check(self, env: Environment) -> None:
        raise NotImplementedError()

    @abstractmethod
    def add_step_callback(self, label: str, cbk: models.TStepCallback) -> None:
        raise NotImplementedError()

    @abstractmethod
    def add_progress_callback(self, label: str, cbk: models.TProgressCallback) -> None:
        raise NotImplementedError()

    @abstractmethod
    def add_learn_callback(self, label: str, cbk: models.TLearnCallback) -> None:
        raise NotImplementedError()

    @abstractmethod
    def call_step_callbacks(self, training_step: models.TrainingStep) -> None:
        raise NotImplementedError()

    @abstractmethod
    def call_progress_callbacks(self, training_progress: models.TrainingProgress) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def call_learn_callbacks(self, learning_step: models.LearningStep) -> None:
        raise NotImplementedError()

    @abstractmethod
    def build_model(self) -> torch.nn.Module:
        raise NotImplementedError()

    @abstractmethod
    def get_self_class_name(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_info(self) -> dict:
        raise NotImplementedError()

    @abstractmethod
    def get_state_dict(self) -> dict:
        raise NotImplementedError()

    @abstractmethod
    def last_layers_model_modifier(self, model: torch.nn.Module) -> torch.nn.Module:
        raise NotImplementedError()

    @abstractmethod
    def agent_specification_model_modifier(self, model: torch.nn.Module) -> torch.nn.Module:
        raise NotImplementedError()

    @abstractmethod
    def model_factory(self) -> torch.nn.Module:
        raise NotImplementedError()

    @abstractmethod
    def preprocess(self, x: T.Any) -> T.Union[torch.Tensor, T.Tuple[torch.Tensor, ...]]:
        raise NotImplementedError()

    @abstractmethod
    def postprocess(self, t: torch.Tensor) -> T.Iterable:
        raise NotImplementedError()

    @abstractmethod
    def act(self, *args) -> T.Iterable:
        raise NotImplementedError()

    @abstractmethod
    def infer(self, preprocessed: T.Union[torch.Tensor, T.Tuple[torch.Tensor, ...]]) -> torch.Tensor:
        raise NotImplementedError()

    @abstractmethod
    def learn(self, entries: T.List[models.ReplayBufferEntry]) -> None:
        raise NotImplementedError()

    @abstractmethod
    def form_learning_batch(self, batch: T.List[models.ReplayBufferEntry]) -> models.LearningBatch:
        raise NotImplementedError()

    @abstractmethod
    def train(self, env: Environment, tp: models.TrainingParams) -> None:
        raise NotImplementedError()
